// --- [ xbee_rx.c ] -------------------------------------------------------------
//
// Ist fr das Empfangen von Frame 0x88, Frame0x8B, Frame0x92, Frame0x97 zustndig
// ber die serielle Schnittstelle eingehende Daten werden zunchst im rx_buffer[]
// abgelegt.
// Sind alle Bytes eines Frames empfangen, wird ein globale Flag gesetzt
// und aus main() knnen dann die Daten des Frames abgeholt werden

#include <avr/io.h>
#include <inttypes.h>
#include <stdlib.h>
#include <avr/interrupt.h>
#include <string.h>

#define FRM_LENGTH          		2			// Bytepositionen innerhalb der
#define FRM_FRAME_TYP          	3			// verschiedenen Frames
#define FRM_FRAME_ID          	4

#define FRM88_STATUS             7
#define FRM88_RESPONSE_START     8

#define FRM8B_DELIVERY_STATUS    8
#define FRM8B_DISCOVERY_STATUS   9

#define FRM90_SOURCE_ADR			12
#define FRM90_RECEIVE_OPTIONS		14
#define FRM90_RX_DATA				15

#define FRM92_SOURCE_ADR         12
#define FRM92_RECEIVE_OPTIONS  	14
#define FRM92_SAMPLES         	15
#define FRM92_DIGITAL_CH_MASK  	16
#define FRM92_ANALOG_CH_MASK   	18
#define FRM92_SAMPLE_START       19

#define FRM97_STATUS             17
#define FRM97_RESPONSE_START     18

#define XBEE_RX_BUFFER_SIZE   	32			// hier muss das grte Frame reinpassen

#include "xbee_rx_a.h"
#include "global.h"
#include "rs232_int.h"

uint16_t FRM_get_word(uint8_t pos);
uint8_t  FRM92_digital_sample(void);		// liefert zurck, ob digitale Samples aktiv

// ----------------------------------------------------------------------------
uint8_t rx_buffer[XBEE_RX_BUFFER_SIZE];	// Empfangspuffer fr ein Frame
volatile uint8_t ISR_rx_idx = 0;				// der Index auf das nchste Byte im Frame

// ----------------------------------------------------------------------------
// speichert ein Frame im Array rx_buffer[]
// sucht nach dem Start Delimiter und nachfolgendem 0
// dann wird das Frameende bestimmt


ISR (USART_RX_vect)
{
static uint8_t frame_end = 0;
register uint8_t rx_in, rx_idx;

rx_idx = ISR_rx_idx;								// volatile in Register einlesen
rx_in = UDR0;                             // empfangenes Byte

rx_buffer[rx_idx] = rx_in;           		// das Byte schon mal speichern
														// aber den Index erst spter incrementieren

if (rx_idx == 0)                         	// das Byte 0 muss 0x7E sein
	{
	if (!(rx_in == START_DELIMITER))
		{
		// es knnte sich um ein Kommando fr die ONE-Wire-Konfiguration handeln
		owi_cmd = rx_in;
		return;										// wenn nicht, dann nichts tun
		}
	}

else if (rx_idx == 1)                     // das Byte 1 muss 0 sein
	{
	if (rx_in)										// wenn Byte 1 doch grer als 0,
		{
		rx_idx = 0;									// dann beenden, dabei Index zurcksetzen
		return;
		}
	}

else if (rx_idx == 2)                     // das Byte 2 ist die Lnge
	{
	frame_end = rx_in + 3;                 // Index des letzten Elementes
	}

else
	{
	if (rx_idx == frame_end)
		{
		rx_idx = 255;								// wird durch incrementieren zu 0
		flag |= (1 << FRM_COMPLETE);			// Flag fr main() setzen
		}
	}

ISR_rx_idx = ++rx_idx;							// Index incrementieren und speichern

TCNT0  = TIMER0_PRELOAD;						// wenn mehr als 4.5 ms keine weiteren
TCCR0B = TIMER0_START;							// Daten ankommen, dann wird ISR_rx_idx auf 0 gesetzt
}


// ----------------------------------------------------------------------------
// Initialiiert den Timer0, aber erst in der ISR (USART_RX_vect) wird er gestartet
void XBEE_rx_timer0_init(void)
{
TCCR0A = 0;
TCCR0B = TIMER0_STOP;
TIMSK0 = (1<<OCIE1A);
}

// ----------------------------------------------------------------------------
// TIMER0_OVF:  3.686400 MHZ / (64 * 256) -> 4.5ms
// setzt ISR_rx_idx zurck, wenn mehr als 4.5 ms keine Aktivitt an der seriellen
// Schnittstelle stattfindet
ISR(TIMER0_COMPA_vect)
{
TCCR0B = TIMER0_STOP;
ISR_rx_idx = 0;									// das ist der Zweck der bung
}

// ----------------------------------------------------------------------------
// liefert den Frame Typ des Frames zurck:
// gleichzeitig wird die Prfsumme getestet.
// Im Fehlerfall wird 0xFF zurckgeliefert, sonst Frame_Type des Frames
uint8_t FRM_get_frame_typ(void)
{
uint8_t i, end, checksum = 0;     			// als 8 Bit-Wert

end = rx_buffer[2] + 3;                 	// Prfsumme einbeziehen
for (i = 3; i <= end; i++) checksum += rx_buffer[i];

if (checksum == 0xFF) return(rx_buffer[FRM_FRAME_TYP]);
else return(0xFF);
}

// --- FRAME 0x88 -------------------------------------------------------------

// ----------------------------------------------------------------------------
// holt den Command Status von Frame0x88
// 0 =  OK, alles andere = Fehler
uint8_t FRM88_get_status(void)
{
return (rx_buffer[FRM88_STATUS]);
}

// ----------------------------------------------------------------------------
// gibt eine Meldung aus mit den Inhalt der Empfangsbesttigung
uint8_t FRM88_get_response(uint8_t * data)
{
uint8_t i = 0, j, end;

end = rx_buffer[FRM_LENGTH] + 3;
j   = FRM88_RESPONSE_START;

while (j < end) data[i++] = rx_buffer[j++];
return(i);
}

// --- FRAME 0x8B -------------------------------------------------------------

// ----------------------------------------------------------------------------
// gibt eine Meldung aus mit den Inhalt der Empfangsbesttigung
uint8_t FRM8B_get_delivery_status(void)
{
return(rx_buffer[FRM8B_DELIVERY_STATUS]);
}

// ----------------------------------------------------------------------------
// gibt eine Meldung aus mit den Inhalt der Empfangsbesttigung
uint8_t FRM8B_get_discovery_status(void)
{
return(rx_buffer[FRM8B_DISCOVERY_STATUS]);
}


// --- FRAME 0x90 -------------------------------------------------------------

// ----------------------------------------------------------------------------
// holt die 16 Bit Netzwerkadresse des Enddevice
// liefert die Anzahl der Bytes zurck

uint8_t FRM90_get_source_adr(uint8_t * data)
{
data[0] = rx_buffer[FRM90_SOURCE_ADR];
data[1] = rx_buffer[FRM90_SOURCE_ADR + 1];

return(2);
}

// ----------------------------------------------------------------------------
// liefert das Receive-Options-Byte des Frames zurck:
// 1 = Acknowledged
// 2 = Broadcast
uint8_t FRM90_get_receive_options(void)
{
return(rx_buffer[FRM90_RECEIVE_OPTIONS]);
}

// ----------------------------------------------------------------------------
// gibt eine Meldung aus mit den Inhalt der Empfangsbesttigung
uint8_t FRM90_get_data(uint8_t * data)
{
uint8_t i = 0, j, end;

end = rx_buffer[FRM_LENGTH] + 3;
j   = FRM90_RX_DATA;

while (j < end) data[i++] = rx_buffer[j++];
return(i);
}

// --- FRAME 0x92 -------------------------------------------------------------

// ----------------------------------------------------------------------------
// liefert das Receive-Options-Byte des Frames zurck:
// 1 = Acknowledged
// 2 = Broadcast
uint8_t FRM92_get_receive_options(void)
{
return(rx_buffer[FRM92_RECEIVE_OPTIONS]);
}

// ----------------------------------------------------------------------------
// holt die 16 Bit Netzwerkadresse des Enddevice
// liefert die Anzahl der Bytes zurck

uint8_t FRM92_get_source_adr(uint8_t * data)
{
data[0] = rx_buffer[FRM92_SOURCE_ADR];
data[1] = rx_buffer[FRM92_SOURCE_ADR + 1];

return(2);
}

// ----------------------------------------------------------------------------
// Prft die digitale IO-Maske:
// wenn kein Bit gesetzt ist, dann wird 0 zurckgeliefert, sonst ein Wert > 0.
// Ist der Wert grer als 0, dann ist mindestens 1 Pin als IO freigeschaltet
// dann wird genau ein Datensatz (2 Byte) mit dem Pinstatus zurckgeliefert.
uint8_t FRM92_digital_sample(void)
{
return (rx_buffer[FRM92_DIGITAL_CH_MASK] + rx_buffer[FRM92_DIGITAL_CH_MASK + 1]);
}

// ----------------------------------------------------------------------------
// liefert der Portstatus als String zurck - zur Vereinfachung hier nur implementiert
// fr die Ports D0 ... D7
//
// Ausgegeben wird in der Form: "_;_;1;0;_;_;1;0;"
// _ = nicht als digitaler Input definiert
// 0 = Pin auf low
// 1 = Pin auf high
// die Pins sind durch Semikolon getrennt
//
// der Status der digitalen Inputs wird in zwei 16Bit Variablen decodiert:
// - die erste Variable zeigt an, ob ein Pin als Input definiert ist (Maske)
// - die zweite Variable zeigt den Status der Pins an
// Die Anzahl der Bytes ist immer gleich 8 * 2 + \0

uint8_t FRM92_get_digital_inputs(uint8_t * txt)
{
uint8_t i, j, m ;
														// wenn kein Bit in der Maske gesetzt ist
														// dann gibt es keine digitalen Daten
if (!(FRM92_digital_sample())) return(0);

j = 8;
i = 0;

while (j)											// hier werden nur die unteren 8Bit getestet !!
   {
	j--;
	m = ( 1<< j );									// Maske bilden
   if (rx_buffer[FRM92_DIGITAL_CH_MASK + 1] & m) // wenn Pin als Input gesetzt
      {
      if (rx_buffer[FRM92_DIGITAL_CH_MASK + 4] & m)  // und High
         {
         txt[i++] = '1';						// Zeichen fr eingeschaltet
         }

      else                                // Pin ist low
         {
         txt[i++] = '0';						// Zeichen fr ausgeschaltet
         }
      }

   else                                   // Pin ist nicht als Input gesetzt
      {
      txt[i++] = '_';							// Zeichen fr deaktiviert
      }

   txt[i++] = ';';                        // Trenner zwischen den Daten
	}

txt[i] = 0;											// die Zeichenfolge mit \0 abschlieen

return(i);                               	// Anzahl der Bytes zurckgeben
}


// ----------------------------------------------------------------------------
// schreibt die analogen Werte in dezimaler Form in einen String,
// die Datenstze sind durch Semikolon getrennt.
// wenn digitale Samples vorhanden sind, dann verschiebt sich das erste analoge Sample
// um 2 Byte nach hinten.
uint8_t FRM92_get_analog_inputs(uint8_t * txt)
{
uint8_t i, end;    								// wenn kein Bit in der Maske gesetzt ist
														// dann liegen keine analogen Samples vor
if ((rx_buffer[FRM92_ANALOG_CH_MASK]) == 0) return(0);
														// Die Samples beginnen ab Pos. 19
														// dort liegen 2 Byte fr digitale Daten
														// sofern digitale Daten gesampled werden
														// Die Position des ersten analogen Samples
														// ist abhngig davon, ob ein digitales Sample vorhanden
if (FRM92_digital_sample()) i = FRM92_SAMPLE_START + 2; else i = FRM92_SAMPLE_START;

														// jeweils 2 Byte bilden ein Sample
end = rx_buffer[FRM_LENGTH] + (2 - 1);		// das Ende des Frames beendet die Samples

txt[0] = 0;											// den String lschen
while(i <= end)
	{
   strcat((char *)txt, utoa( (rx_buffer[i] * 256 + rx_buffer[i+1]), msg, 10));
   strcat((char *)txt, ";");					// Samples durch Semikolon trennen
	i += 2;
   }
return(strlen((char*) txt));
}

// --- FRAME 0x97 -------------------------------------------------------------

// ----------------------------------------------------------------------------
// liefert den Command Status binr zurck. Nur 0 ist ein Erfolg.
uint8_t FRM97_get_status(void)
{
return(rx_buffer[FRM97_STATUS]);
}

// ----------------------------------------------------------------------------
// schreibt die empfangenen Bytes in das Array data[] und
// liefert die Anzahl der empfangenen Bytes zurck
uint8_t FRM97_get_response(uint8_t * data)
{
uint8_t i = 0, j, end;

end = rx_buffer[FRM_LENGTH] + 3;
j   = FRM97_RESPONSE_START;

while (j < end) data[i++] = rx_buffer[j++];
return(i);
}

/*
// ----------------------------------------------------------------------------
// zum Debuggen
void FRM_debug_print(void)
{
uint8_t i, end;

end = rx_buffer[2] + 3;
put_c(13);
for (i = 0; i <= end; i++)
	{
	put_s(utoa( rx_buffer[i], msg, 16));
	put_c(0);							// als Trenner zwischen den Daten

	}
put_c(13);
}
*/
// --- [ eof ] ----------------------------------------------------------------
